/*
 * Strawberry Music Player
 * Copyright 2010, David Sansome <me@davidsansome.com>
 * Copyright 2018-2021, Jonas Kvinge <jonas@jkvinge.net>
 *
 * Strawberry is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Strawberry is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Strawberry.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <QString>

#include "scopedwchararray.h"

ScopedWCharArray::ScopedWCharArray(const QString &str)
    // Allocate from str.length() rather than chars_ so the allocation size does not depend on member-initialization order (data_ must not read chars_ before it is set).
    : chars_(str.length()), data_(new wchar_t[static_cast<size_t>(str.length()) + 1]) {
  str.toWCharArray(data_.get());
  data_[static_cast<size_t>(chars_)] = '\0';
}
